"""Implement some positive semidefinite kernels."""

from numpy import array, matrix
from numpy import zeros, ones, eye, diag
from numpy import average, exp, kron, sum
from numpy import sqrt, divide, mean, multiply

try:
    from kernelbase import Kernel, JointKernel
except ImportError:
    class Kernel(object):
        """A Mercer kernel. Note that if Cython is available, then this is not used."""
        def __init__(self):
            self.kmat = array([])
            self.examples = None

        def __getitem__(self, idx):
            assert(len(idx)==2)
            return self.kmat[idx[0], idx[1]]

        def __repr__(self):
            rep = '<' + self.__class__.__name__ + ' instance>'
            return rep

        def dump(self) :
            """
            returns a string that can be used to construct an equivalent object
            """
            kstr = self.__module__ + '.' + self.__class__.__name__ + '(' + \
                   self.construction_params() + ')'
            return kstr

        def construction_params(self):
            raise NotImplementedError

        def compute(self, examples):
            """Evaluate the kernel function and store the value"""
            raise NotImplementedError

        def compute_test(self, test_ex):
            """Evaluate the kernel function on test examples"""
            raise NotImplementedError


    class JointKernel(object):
        """A kernel on input and output pairs."""
        def __init__(self, input_kernel, output_kmat=None, num_classes=-1):
            self.Kx = input_kernel
            if output_kmat:
                self.Ky = output_kmat
            else:
                assert(num_classes > 1)
                self.Ky = eye(num_classes)
            self.examples = None

        def __repr__(self) :
            rep = '<' + self.__class__.__name__ + ' instance>'
            return rep

        def __getitem__(self, idx):
            """Simulate matrix element.
            Order is assumed to be examples first then data.
            """
            assert(len(idx)==4)
            return self.Kx[idx[0], idx[1]]*self.Ky[idx[2], idx[3]]

        @property
        def kmat(self):
            """Simulate the whole kernel matrix"""
            return kron(self.Kx.kmat, self.Ky)

        def compute(self, examples):
            self.examples = examples
            self.Kx.compute(examples)

        def compute_test(self, test_ex):
            """Returns a non-square matrix, of num_train by num_test size."""
            Kx = self.Kx.compute_test(test_ex)
            return kron(Kx, self.Ky)

class LinearKernel(Kernel):
    """Standard dot product"""
    def __init__(self):
        Kernel.__init__(self)
        
    def construction_params(self):
        return ''
    
    def compute(self, examples):
        """Evaluate the kernel function and store the value"""
        self.examples = examples
        self.kmat = array(matrix(examples).T*matrix(examples))

    def compute_test(self, test_ex):
        """Evaluate the kernel function on test examples"""
        if test_ex.ndim == 1:
            test_ex.shape = (len(test_ex),1)
        return array(matrix(self.examples).T*matrix(test_ex))
    

class GaussKernel(Kernel):
    """Gaussian rbf kernel"""
    def __init__(self, gamma=1.0):
        Kernel.__init__(self)
        self.gamma = gamma

    def __repr__(self) :
        rep = '<' + self.__class__.__name__ + ' instance>\n'
        rep += 'gamma : ' + str(self.gamma)
        return rep

    def constructionParams(self) :
        return 'gamma = ' + str(self.gamma)

    def compute(self, examples):
        """Evaluate the kernel function and store the value"""
        self.examples = examples
        self.kmat = exp(-self.gamma * sqr_dist(examples, examples))

    def compute_test(self, test_ex):
        """Evaluate the kernel function on test examples"""
        if test_ex.ndim == 1:
            test_ex.shape = (len(test_ex),1)
        return exp(-self.gamma * sqr_dist(self.examples, test_ex))
    

class CustomKernel(Kernel):
    """Just a positive semidefinite matrix"""
    def __init__(self):
        Kernel.__init__(self)

    def compute(self, examples):
        """Assume input is a kernel matrix.
        Copy the training kernel matrix"""
        self.kmat = examples

    def compute_test(self, test_ex):
        """Assume input is a kernel matrix.
        Just return the test kernel"""
        return test_ex

#####################################################################
# Some miscellaneaous functions for kernel matrix operations
#####################################################################

def sqr_dist(a,b):
    """Compute the square distance between vectors"""
    dot_a = sum(a*a, axis=0).T
    dot_b = sum(b*b ,axis=0).T
    unitvec = ones(dot_a.shape)
    D = array(2.0*matrix(a).T*matrix(b))
    
    for ix,bval in enumerate(dot_b):
        D[:,ix] = dot_a - D[:,ix] + kron(bval,unitvec)

    return D

def center(K):
    """Center the kernel matrix, such that the mean (in feature space) is zero."""
    one_mat = matrix(ones(K.shape))
    one_vec = matrix(ones((K.shape[0],1)))

    row_sum = matrix(mean(K,axis=0)).T
    R = K - row_sum * one_vec.T - one_vec * row_sum.T +\
        mean(row_sum.A)*one_mat
    return R

def normalize_variance(K):
    """Normalize the kernel matrix, such that the variance (in feature space) is 1"""
    one_vec = matrix(ones((K.shape[0],1)))
    inv_sqrt_diag = divide(one_vec, matrix(sqrt(diag(K))).T)
    KN = multiply(kron(one_vec.T,inv_sqrt_diag),K)
    KN = multiply(kron(one_vec,inv_sqrt_diag.T),K)
    return KN

def normalize_unit_diag(K):
    """Normalize the kernel matrix, such that all diagonal entries are 1."""
    Kii = diag(K)
    Kii.shape = (len(Kii),1)
    return divide(K, sqrt(matrix(Kii)*matrix(Kii).T))

def kernel2distance(K):
    """Convert the kernel matrix into the corresponding Euclidean distance."""
    # TODO: Vectorize
    D = zeros(K.shape)
    for ix in range(K.shape[0]):
        for iy in range(K.shape[1]):
            D[ix,iy] = sqrt(K[ix,ix] + K[iy,iy] - 2*K[ix,iy])
    return D

#####################################################################
# Some functions for testing
#####################################################################

def test_kernel(kfunc, X):
    print '=== Test Kernel ==='
    kg = kfunc()
    print kg
    kg.compute(X)
    print kg.kmat
    for ix in range(X.shape[1]):
        for iy in range(X.shape[1]):
            print '(%d,%d) %f' % (ix, iy, kg[ix,iy])
            
def test_joint_kernel(kfunc, X):
    print '=== Test Joint Kernel ==='
    joint = JointKernel(kfunc(), num_classes=2)
    print joint
    print joint.Kx
    joint.compute(X)
    print joint.kmat
    for xi in range(X.shape[1]):
        for yi in range(2):
            for xj in range(X.shape[1]):
                for yj in range(2):
                    print 'Kx(%d,%d),Ky(%d,%d): %f' % (xi,xj,yi,yj, joint[xi,xj,yi,yj])
    
def test_prediction(kfunc, Xtrain, Xtest):
    print '=== Test prediction ==='
    kg = kfunc()
    print kg
    kg.compute(Xtrain)
    print kg.compute_test(Xtest)
    joint = JointKernel(kfunc(), num_classes=2)
    joint.compute(Xtrain)
    print joint.compute_test(Xtest)
    
if __name__ == '__main__':
    from numpy.random import rand
    X = rand(2,3)
    test_kernel(LinearKernel, X)
    test_kernel(GaussKernel, X)
    test_joint_kernel(LinearKernel, X)
    test_joint_kernel(GaussKernel, X)
    Xtest = rand(2,1)
    test_prediction(LinearKernel, X, Xtest)
    test_prediction(GaussKernel, X, Xtest)


