"""Implement a Cython interface for getitem."""

from numpy import array
import numpy as np
cimport numpy as np
np.import_array()
cimport cython

DTYPE = np.float
ctypedef np.float_t DTYPE_t


@cython.boundscheck(False) # turn of bounds-checking for entire function
def getitem2(np.ndarray[DTYPE_t, ndim=2, negative_indices=False] kmat, Py_ssize_t xi, Py_ssize_t xj):
    """Fast version of __getitem__ with 2 arguments"""
    return <float> kmat[xi, xj]

@cython.boundscheck(False) # turn of bounds-checking for entire function
def getitem4(np.ndarray[DTYPE_t, ndim=2, negative_indices=False] Kx,
             np.ndarray[DTYPE_t, ndim=2, negative_indices=False] Ky,
             Py_ssize_t xi, Py_ssize_t xj, Py_ssize_t yi, Py_ssize_t yj):
    """Fast version of __getitem__ with 4 arguments"""
    return <float> Kx[xi, xj]*Ky[yi, yj]

class Kernel(object):
    """A Mercer kernel, with Cython interface for getitem"""
    def __init__(self):
        self.kmat = array([],dtype=DTYPE)
        self.examples = None
        
    def __getitem__(self, idx):
        assert(len(idx)==2)
        (xi,xj) = idx
        return getitem2(self.kmat, xi, xj)

    def __repr__(self):
        rep = '<' + self.__class__.__name__ + ' instance>'
        return rep

    def dump(self) :
        """
        returns a string that can be used to construct an equivalent object
        """
        kstr = self.__module__ + '.' + self.__class__.__name__ + '(' + \
               self.construction_params() + ')'
        return kstr

    def construction_params(self):
        raise NotImplementedError

    def compute(self, examples):
        """Evaluate the kernel function and store the value"""
        raise NotImplementedError

    def compute_test(self, test_ex):
        """Evaluate the kernel function on test examples"""
        raise NotImplementedError


class JointKernel(object):
    """A kernel on input and output pairs."""
    def __init__(self, input_kernel, output_kmat=None, num_classes=-1):
        self.Kx = input_kernel
        if output_kmat:
            self.Ky = output_kmat
        else:
            assert(num_classes > 1)
            self.Ky = np.eye(num_classes)
        self.examples = None

    def __repr__(self) :
        rep = '<' + self.__class__.__name__ + ' instance>'
        return rep

    def __getitem__(self, idx):
        """Simulate matrix element.
        Order is assumed to be examples first then data.
        """
        assert(len(idx)==4)
        (xi, xj, yi, yj) = idx
        return getitem4(self.Kx.kmat, self.Ky, xi, xj, yi, yj)
    
    @property
    def kmat(self):
        """Simulate the whole kernel matrix"""
        return np.kron(self.Kx.kmat, self.Ky)

    def compute(self, examples):
        self.examples = examples
        self.Kx.compute(examples)

    def compute_test(self, test_ex):
        """Returns a non-square matrix, of num_train by num_test size."""
        Kx = self.Kx.compute_test(test_ex)
        return np.kron(Kx, self.Ky)

