"""Some functions to help test the modules"""
import pylab
from os import link
from numpy import ones, concatenate, log, exp
from numpy.random import randn, rand, permutation, randint, seed
from numpy import linspace, where, sqrt, square, median
from numpy import zeros, eye, mean, kron, sort, sign
from numpy import array, nonzero, equal, ix_
from optwok.mldata import DatasetFileCSV
from optwok.kernel import sqr_dist

def cloudgen(numpoint, numfeat, numnoise, fracpos, width):
    """Generate two Gaussian point clouds, centered around one and minus one.
    Gaussian clouds are in numfeat dimensions.
    Generate uniform noise in numnoise dimensions.
    """
    numpos = int(round(fracpos*numpoint))
    numneg = numpoint - numpos

    metadata = 'cloudgen(%d,%d,%d,%d,%3.2f)' % (numpos, numneg, numfeat, numnoise, width)
    print metadata

    datapos = ones((numfeat, numpos)) + width*randn(numfeat, numpos)
    dataneg = -ones((numfeat, numneg)) + width*randn(numfeat, numneg)
    noise = (2.0+width)*(rand(numnoise, numpos+numneg)
                             -0.5*ones((numnoise, numpos+numneg)))
    #noise = sign(rand(numnoise, numpos+numneg)-0.5*ones((numnoise, numpos+numneg)))\
    #       + width*randn(numnoise, numpos+numneg)
    pointcloud = 0.2*concatenate((concatenate((datapos,dataneg),axis=1),
                                  noise), axis=0)
    labels = concatenate((ones(numpos),-ones(numneg)))

    return metadata, pointcloud, labels

def cloudgen_weak(numpoint, numfeat, numweak, fracpos, width):
    """Generate two Gaussian point clouds, centered around one and minus one.
    Gaussian clouds are in numfeat dimensions.
    Overlapping Gaussian clouds are in numweak dimensions.
    """
    numpos = int(round(fracpos*numpoint))
    numneg = numpoint - numpos

    metadata = 'cloudgen_weak(%d,%d,%d,%d,%3.2f)' %\
               (numpos, numneg, numfeat, numweak, width)
    print metadata

    strongpos = ones((numfeat, numpos)) + width*randn(numfeat, numpos)
    strongneg = -ones((numfeat, numneg)) + width*randn(numfeat, numneg)
    weakpos = 0.1*ones((numweak, numpos)) + 2.0*width*randn(numweak, numpos)
    weakneg = -0.1*ones((numweak, numneg)) + 2.0*width*randn(numweak, numneg)
    datapos = concatenate((strongpos, weakpos), axis=0)
    dataneg = concatenate((strongneg, weakneg), axis=0)
    pointcloud = 0.2*concatenate((datapos,dataneg),axis=1)
    labels = concatenate((ones(numpos),-ones(numneg)))

    return metadata, pointcloud, labels

def multi_cloudgen(num_point, num_feat, num_class, edge_length):
    """Generate numclass number of Gaussian point clouds,
    centered around the simplex of numclass."""
    assert(num_feat >= num_class)
    centers = zeros((num_feat, num_class))
    centers[:num_class, :num_class] = (edge_length/sqrt(2))*eye(num_class)
    midpoint = mean(centers,axis=1)
    midpoint.shape = (num_feat,1)
    centers = centers - kron(ones((1,num_class)),midpoint)

    if type(num_point) == type(0):
        num_point_list = []
        for ix in range(num_class):
            num_point_list.append(num_point)
        num_point = num_point_list
    metadata = 'multi_cloudgen(%s,%d,%d,%3.2f)' % (num_point, num_feat, num_class, edge_length)

    pointcloud = zeros((num_feat,0))
    labels = array([],dtype=int)
    for ix in range(num_class):
        cur_cloud = kron(ones((num_point[ix],1)),centers[:,ix]).T + randn(num_feat, num_point[ix])
        pointcloud = concatenate((pointcloud, cur_cloud), axis=1)
        labels = concatenate((labels, ix*ones(num_point[ix])))

    return metadata, pointcloud, labels

def multi_cloudgen_mix(num_point, num_feat, num_class, edge_length, mix_list):
    """Generate num_class Gaussian point clouds,
    but labels are mixed depending on mix_list.
    
    """
    (metadata, pointcloud, labels) = multi_cloudgen(num_point, num_feat, num_class, edge_length)
    metadata += str(mix_list)
    new_labels = zeros(labels.shape, dtype=int)
    if len(mix_list) == num_class:
        # split classes
        for cid in range(num_class):
            idx = nonzero(equal(labels,cid))[0]
            if len(mix_list[cid]) > 1:
                new_labels[idx] = randint(min(mix_list[cid]), max(mix_list[cid])+1, len(idx))
            else:
                new_labels[idx] = mix_list[cid][0]
    else:
        # merge classes
        assert(len(flatten_list(mix_list)) == num_class)
        for cid in range(len(mix_list)):
            idx = array([], dtype=int)
            for mix in mix_list[cid]:
                idx = concatenate([idx, nonzero(equal(labels, mix))[0]])
            new_labels[idx] = cid
    return metadata, pointcloud, new_labels
        
def plot_confusion(mix_list):
    """Plot the kernel matrix corresponding to a particular
    mix_list denoting the mixing of labels."""
    k = len(flatten_list(mix_list))
    Ky = zeros((k,k))
    for block in mix_list:
        Ky[ix_(block,block)] = 1.0
    pylab.matshow(Ky)
    pylab.colorbar()

def random_classification_noise(labels, frac_flip, symm_flip):
    """Generate random classification noise by flipping a proportion
    of labels randomly.
    """
    if frac_flip>0.0:
        num_ex = len(labels)
        if symm_flip:
            flip_idx = permutation(num_ex)[:round(frac_flip*num_ex)]
            labels[flip_idx] = -1.0*labels[flip_idx]
        else:
            flip_idx = permutation(num_ex)[:round(2.0*frac_flip*num_ex)]
            labels[flip_idx] = -1.0*labels[where(labels[flip_idx]>0.0)]

def create_data(num_train=500, num_test=200, num_feat=10, num_noise_feat=10, width=0.8,
                frac_pos=0.5, frac_flip=0.0, symm_flip=True):
    """Create mixture of Gaussians distribution"""
    (metadata, train_ex, train_labels) = cloudgen(num_train, num_feat,
                                                  num_noise_feat, frac_pos, width)
    (metadata, test_ex, test_labels) = cloudgen(num_test, num_feat,
                                                num_noise_feat, frac_pos, width)
    random_classification_noise(train_labels, frac_flip, symm_flip)
    
    return (train_ex, train_labels, test_ex, test_labels)

def plot_2d(x, y, svmout, train_ex, train_lab, axis_limit, title):
    """Plot 2D data"""
    pylab.figure()
    pylab.pcolor(x, y, svmout)
    pylab.contour(x, y, svmout, linewidths=1, colors='black', hold=True)
    pos_lab = train_lab > 0.0
    neg_lab = train_lab < 0.0
    pylab.plot(train_ex[0,pos_lab], train_ex[1,pos_lab], 'r+')
    pylab.plot(train_ex[0,neg_lab], train_ex[1,neg_lab], 'bx')
    pylab.axis(axis_limit)
    pylab.title(title)

def twoballs_filename(data_dir, num_examples, num_feat, num_noise_feat, frac_flip):
    """Generate filename to save data and permutations"""
    data_filename = data_dir + '/twoballs_n=%d_%d:%d_rcn=%1.1f.csv'\
             % (num_examples, num_feat, num_noise_feat, frac_flip)
    perm_filename = data_dir + '/twoballs_n=%d_%d:%d_rcn=%1.1f_perm.txt'\
                    % (num_examples, num_feat, num_noise_feat, frac_flip)
    return (data_filename, perm_filename)

def save_cloudgen(data_dir, num_noise_feat, frac_flip, symm_flip=True,
                  num_ex=100, num_feat=10, frac_pos=0.5, width=1.5, num_perms=50):
    """Generate two gaussians, with noise dimensions and
    random classification noise, and save the dataset.
    """
    assert(frac_flip>=0.0 and frac_flip<=0.5)
    (metadata, examples, labels) = cloudgen(num_ex, num_feat, num_noise_feat,
                                            frac_pos, width)
    random_classification_noise(labels, frac_flip, symm_flip)
    (data_filename, perm_filename) = twoballs_filename(data_dir, num_ex, num_feat, num_noise_feat, frac_flip)
    outfile = DatasetFileCSV(data_filename, 'vec', verbose=True)
    outfile.writelines(examples, labels)

    handle = open(perm_filename, 'w')
    for iperm in range(num_perms):
        cur_mix = permutation(num_ex)
        handle.write(' '.join(map(str, cur_mix.tolist())))
        handle.write('\n')
    handle.close()

def sparse_data(data_dir):
    """Generate data to investigate sparsity,
    with respect to both features and support vectors."""
    seed(1)
    num_feat = 10
    num_examples = [50, 100]
    noise_feats = [0, 10, 100, 200]
    rcns = [0.0, 0.1, 0.2, 0.3]

    num_datasets = len(noise_feats)
    assert(len(rcns)==num_datasets)

    for num_ex in num_examples:
        for num_noise_feat in noise_feats:
            save_cloudgen(data_dir, num_noise_feat, 0.0, num_ex=num_ex, num_feat=num_feat)
        (data_filename, perm_filename) = twoballs_filename(data_dir, num_ex, num_feat, 0, 0.0)
        # Add random classification noise in the training
        for rcn in rcns[1:]:
            (df, pf) = twoballs_filename(data_dir, num_ex, num_feat, 0, rcn)
            link(data_filename, df)
            link(perm_filename, pf)
        for ix in range(1,num_datasets):
            (dfs, pfs) = twoballs_filename(data_dir, num_ex, num_feat, noise_feats[ix], 0.0)
            (df, pf) = twoballs_filename(data_dir, num_ex, num_feat, noise_feats[ix], rcns[ix])
            link(dfs, df)
            link(pfs, pf)

def flatten_list(general_list):
    """Convert a list of lists into one list. Works recursively."""
    if isinstance(general_list, list):
        return sum((flatten_list(x) for x in general_list), [])
    else:
        return [general_list]
    
def get_gauss_widths(examples, num_widths, start_quantile=0.1, stop_quantile=0.9):
    """Return a candidate set of possible widths
    for use in the Gaussian kernel.
    """
    distance = sqr_dist(examples, examples)
    # get the sorted distances (ignoring distances to self)
    dm = sort(distance.flatten())
    dm = square(dm[examples.shape[1]:])
    if num_widths == 1:
        widths = median(dm)
    else:
        # find points uniformly spaces in log scale
        start_idx = round(start_quantile*len(dm))
        stop_idx = round(stop_quantile*len(dm))
        dm = log(dm)
        widths = exp(linspace(dm[stop_idx],dm[start_idx],num_widths))
    print widths
    return widths
